<?php
/**
 * Manager to help with getting values from a configuration array.
 *
 * @copyright   2013-2019 Perforce Software. All rights reserved.
 * @license     Please see LICENSE.txt in top-level readme folder of this distribution.
 * @version     2019.2/1886137
 */

namespace Application\Config;

use Users\Settings\ReviewPreferences;
use Users\Settings\TimePreferences;
use Reviews\Model\Review;
use Workflow\Model\IWorkflow;
use Zend\ModuleManager\Listener\ConfigMergerInterface;
use Markdown\Settings as MarkdownSettings;
use Zend\ModuleManager\ModuleEvent;
use ReflectionClass;
use ReflectionException;
use Exception;
use InvalidArgumentException;

class ConfigManager
{
    const HOURS_MINUTES_24_PATTERN                = '/^(([01][0-9]|2[0-3]):([0-5][0-9]))$/';
    const CONFIG                                  = 'config';
    const ENVIRONMENT                             = 'environment';
    const LOGOUT_URL                              = 'logout_url';
    const BASE_URL                                = 'base_url';
    const ASSET_BASE_PATH                         = 'asset_base_path';
    const VENDOR                                  = 'vendor';
    const EMOJI_PATH                              = 'emoji_path';
    const ENVIRONMENT_LOGOUT_URL                  = 'environment.logout_url';
    const ENVIRONMENT_EMOJI_URL                   = 'environment.vendor.emoji_path';
    const ENVIRONMENT_BASE_URL                    = self::ENVIRONMENT . '.' . self::BASE_URL;
    const ENVIRONMENT_ASSET_BASE_PATH             = self::ENVIRONMENT . '.' . self::ASSET_BASE_PATH;
    const P4                                      = 'p4';
    const SSO_ENABLED                             = 'sso_enabled';
    const P4_SSO_ENABLED                          = 'p4.sso_enabled';
    const DIFF_MAX_DIFFS                          = 'diffs.max_diffs';
    const FILES_DOWNLOAD_TIMEOUT                  = 'files.download_timeout';
    const FILES_MAX_SIZE                          = 'files.max_size';
    const EXPAND_ALL_FILE_LIMIT                   = 'expand_all_file_limit';
    const REVIEWS_EXPAND_ALL                      = self::REVIEWS . '.' . self::EXPAND_ALL_FILE_LIMIT;
    const DISABLE_TESTS_ON_APPROVE_COMMIT         = 'disable_tests_on_approve_commit';
    const REVIEWS_DISABLE_TESTS_ON_APPROVE_COMMIT = self::REVIEWS . '.' . self::DISABLE_TESTS_ON_APPROVE_COMMIT;
    const DISABLE_APPROVE_WHEN_TASKS_OPEN         = 'disable_approve_when_tasks_open';
    const REVIEWS_DISABLE_APPROVE_WHEN_TASKS_OPEN = self::REVIEWS . '.' . self::DISABLE_APPROVE_WHEN_TASKS_OPEN;
    const EXPAND_GROUP_REVIEWERS                  = 'expand_group_reviewers';
    const REVIEWS_EXPAND_GROUP_REVIEWERS          = self::REVIEWS . '.' . self::EXPAND_GROUP_REVIEWERS;
    const FILTERS                                 = 'filters';
    const RESULT_SORTING                          = 'result_sorting';
    const REVIEWS_FILTERS_RESULT_SORTING          = self::REVIEWS . '.' . self::FILTERS . '.' . self::RESULT_SORTING;
    const DATE_FIELD                              = 'date_field';
    const REVIEWS_FILTERS_DATE_FIELD              = self::REVIEWS . '.'. self::FILTERS . '.' . self::DATE_FIELD;
    const UPGRADE_BATCH_SIZE                      = 'upgrade.batch_size';
    const UPGRADE_STATUS_REFRESH_INTERVAL         = 'upgrade.status_refresh_interval';
    const AVATARS_HTTP                            = 'avatars.http_url';
    const AVATARS_HTTPS                           = 'avatars.https_url';
    const SECURITY_REQUIRE_LOGIN                  = 'security.require_login';
    const QUEUE_WORKER_CHANGE_SAVE_DELAY          = self::QUEUE . '.' . self::WORKER_CHANGE_SAVE_DELAY;
    const TRANSLATOR_NON_UTF8_ENCODINGS           = 'translator.non_utf8_encodings';
    const DASHBOARD_MAX_ACTIONS                   = 'users.maximum_dashboard_actions';
    const DASHBOARD_REFRESH_INTERVAL              = 'users.dashboard_refresh_interval';
    const DISPLAY_FULLNAME                        = 'users.display_fullname';
    const README_MODE                             = 'readme_mode';
    const PROJECTS_README_MODE                    = self::PROJECTS . '.' . self::README_MODE;
    const MAX_README_SIZE                         = 'max_readme_size';
    const PROJECTS_MAX_README_SIZE                = self::PROJECTS . '.' . self::MAX_README_SIZE;
    const FETCH                                   = 'fetch';
    const MAXIMUM                                 = 'maximum';
    const PROJECTS_FETCH_MAXIMUM                  = self::PROJECTS . '.'.self::FETCH.'.'. self::MAXIMUM;
    const MAINLINES                               = 'mainlines';
    const PROJECTS_MAINLINES                      = self::PROJECTS . '.' . self::MAINLINES;
    const COMMENT_SHOW_ID                         = 'comments.show_id';
    const COMMENT_THREADING_MAX_DEPTH             = 'comments.threading.max_depth';
    const COMMENT_NOTIFICATION_DELAY_TIME         = 'comments.notification_delay_time';
    const JIRA                                    = 'jira';
    const RELATIONSHIP                            = 'relationship';
    const LINK_TO_JOBS                            = 'link_to_jobs';
    const DELAY_JOB_LINKS                         = 'delay_job_links';
    const MAX_JOB_FIXES                           = 'max_job_fixes';
    const JOB_FIELD                               = 'job_field';
    const HOST                                    = 'host';
    const USER                                    = 'user';
    const PASSWORD                                = 'password';
    const JIRA_LINK_TO_JOBS                       = self::JIRA . '.' . self::LINK_TO_JOBS;
    const JIRA_DELAY_JOB_LINKS                    = self::JIRA . '.' . self::DELAY_JOB_LINKS;
    const JIRA_MAX_JOB_FIXES                      = self::JIRA . '.' . self::MAX_JOB_FIXES;
    const JIRA_HOST                               = self::JIRA . '.' . self::HOST;
    const JIRA_USER                               = self::JIRA . '.' . self::USER;
    const JIRA_PASSWORD                           = self::JIRA . '.' . self::PASSWORD;
    const JIRA_JOB_FIELD                          = self::JIRA . '.' . self::JOB_FIELD;
    const JIRA_RELATIONSHIP                       = self::JIRA . '.' . self::RELATIONSHIP;
    const PROCESS_SHELF_DELETE_WHEN               = 'process_shelf_delete_when';
    const REVIEWS_PROCESS_SHELF_DELETE_WHEN       = self::REVIEWS . '.' . self::PROCESS_SHELF_DELETE_WHEN;
    const MORE_CONTEXT_LINES                      = 'more_context_lines';
    const REVIEWS_MORE_CONTEXT_LINES              = self::REVIEWS . '.' . self::MORE_CONTEXT_LINES;
    const MAX_BOTTOM_CONTEXT_LINES                = 'max_bottom_context_lines';
    const REVIEWS_MAX_BOTTOM_CONTEXT_LINES        = self::REVIEWS . '.' . self::MAX_BOTTOM_CONTEXT_LINES;
    const ALLOW_AUTHOR_OBLITERATE                 = 'allow_author_obliterate';
    const REVIEWS_ALLOW_AUTHOR_OBLITERATE         = self::REVIEWS . '.' . self::ALLOW_AUTHOR_OBLITERATE;
    const DISABLE_SELF_APPROVE                    = 'disable_self_approve';
    const REVIEWS_DISABLE_SELF_APPROVE            = self::REVIEWS . '.' . self::DISABLE_SELF_APPROVE;
    const DISABLE_COMMIT                          = 'disable_commit';
    const REVIEWS_DISABLE_COMMIT                  = self::REVIEWS . '.' . self::DISABLE_COMMIT;
    const MODERATOR_APPROVAL                      = 'moderator_approval';
    const REVIEWS_MODERATOR_APPROVAL              = self::REVIEWS . '.' . self::MODERATOR_APPROVAL;
    const END_STATES                              = 'end_states';
    const REVIEWS_END_STATES                      = self::REVIEWS . '.' . self::END_STATES;
    const SECURITY_PREVENT_LOGIN                  = 'security.prevent_login';
    const SECURITY_HTTPS_STRICT                   = 'security.https_strict';
    const MENTIONS                                = 'mentions';
    const USERS_BLACKLIST                         = 'usersBlacklist';
    const GROUPS_BLACKLIST                        = 'groupsBlacklist';
    const MODE                                    = 'mode';
    const MENTIONS_USERS_BLACKLIST                = self::MENTIONS . '.' . self::USERS_BLACKLIST;
    const MENTIONS_GROUPS_BLACKLIST               = self::MENTIONS . '.' . self::GROUPS_BLACKLIST;
    const MENTIONS_MODE                           = self::MENTIONS . '.' . self::MODE;
    const LOG                                     = 'log';
    const REFERENCE_ID                            = 'reference_id';
    const LOG_REFERENCE_ID                        = self::LOG . '.' . self::REFERENCE_ID;
    const EVENT_TRACE                             = 'event_trace';
    const LOG_EVENT_TRACE                         = self::LOG . '.' . self::EVENT_TRACE;
    const MARKDOWN                                = 'markdown';
    const FILE_EXTENSIONS                         = 'file_extensions';
    const MARKDOWN_FILE_EXTENSIONS                = self::MARKDOWN . '.' . self::FILE_EXTENSIONS;
    const MARKDOWN_MARKDOWN                       = self::MARKDOWN . '.' . self::MARKDOWN;
    const FILE                                    = 'file';
    const LOG_FILE                                = self::LOG . '.' . self::FILE;
    const PRIORITY                                = 'priority';
    const LOG_PRIORITY                            = self::LOG . '.' . self::PRIORITY;
    const DEPOT_STORAGE                           = 'depot_storage';
    const BASE_PATH                               = 'base_path';
    const DEPOT_STORAGE_BASE_PATH                 = self::DEPOT_STORAGE . '.' . self::BASE_PATH;
    const SECURITY                                = 'security';
    const ADD_PROJECT_ADMIN_ONLY                  = 'add_project_admin_only';
    const PROJECTS                                = 'projects';
    const ADD_ADMIN_ONLY                          = 'add_admin_only';
    const ADD_GROUPS_ONLY                         = 'add_groups_only';
    const ADD_PROJECT_GROUPS                      = 'add_project_groups';
    const SECURITY_ADD_PROJECT_ADMIN_ONLY         = self::SECURITY . '.' . self::ADD_PROJECT_ADMIN_ONLY;
    const SECURITY_ADD_PROJECT_GROUPS             = self::SECURITY . '.' . self::ADD_PROJECT_GROUPS;
    const PROJECTS_ADD_ADMIN_ONLY                 = self::PROJECTS . '.' . self::ADD_ADMIN_ONLY;
    const PROJECTS_ADD_GROUPS_ONLY                = self::PROJECTS . '.' . self::ADD_GROUPS_ONLY;
    const QUEUE                                   = 'queue';
    const PATH                                    = 'path';
    const WORKERS                                 = 'workers';
    const WORKER_LIFETIME                         = 'worker_lifetime';
    const WORKER_TASK_TIMEOUT                     = 'worker_task_timeout';
    const WORKER_MEMORY_LIMIT                     = 'worker_memory_limit';
    const DISABLE_TRIGGER_DIAGNOSTICS             = 'disable_trigger_diagnostics';
    const QUEUE_PATH                              = self::QUEUE . '.' . self::PATH;
    const QUEUE_WORKERS                           = self::QUEUE . '.' . self::WORKERS;
    const QUEUE_WORKER_LIFETIME                   = self::QUEUE . '.' . self::WORKER_LIFETIME;
    const QUEUE_WORKER_TASK_TIMEOUT               = self::QUEUE . '.' . self::WORKER_TASK_TIMEOUT;
    const QUEUE_WORKER_MEMORY_LIMIT               = self::QUEUE . '.' . self::WORKER_MEMORY_LIMIT;
    const QUEUE_DISABLE_TRIGGER_DIAGNOSTICS       = self::QUEUE . '.' . self::DISABLE_TRIGGER_DIAGNOSTICS;
    const WORKER_CHANGE_SAVE_DELAY                = 'worker_change_save_delay';
    const REVIEWS                                 = 'reviews';
    const SYNC_DESCRIPTIONS                       = 'sync_descriptions';
    const REVIEWS_SYNC_DESCRIPTIONS               = self::REVIEWS . '.' . self::SYNC_DESCRIPTIONS;
    const EMULATE_IP_PROTECTIONS                  = 'emulate_ip_protections';
    const SECURITY_EMULATE_IP_PROTECTIONS         = self::SECURITY . '.' . self::EMULATE_IP_PROTECTIONS;
    const REDIS                                   = 'redis';
    const OPTIONS                                 = 'options';
    const NAMESPACE                               = 'namespace';
    const REDIS_OPTIONS_NAMESPACE                 = self::REDIS . '.' . self::OPTIONS . '.' . self::NAMESPACE;
    const MAIL                                    = 'mail';
    const VALIDATOR                               = 'validator';
    const MAIL_VALIDATOR_OPTIONS                  = self::MAIL . '.' . self::VALIDATOR . '.' . self::OPTIONS;
    const POPULATION_LOCK_TIMEOUT                 = 'population_lock_timeout';
    const CHECK_INTEGRITY                         = 'check_integrity';
    const ITEMS_BATCH_SIZE                        = 'items_batch_size';
    const REDIS_POPULATION_LOCK_TIMEOUT           = self::REDIS . '.' . self::POPULATION_LOCK_TIMEOUT;
    const REDIS_CHECK_INTEGRITY                   = self::REDIS . '.' . self::CHECK_INTEGRITY;
    const REDIS_ITEMS_BATCH_SIZE                  = self::REDIS . '.' . self::ITEMS_BATCH_SIZE;
    const ALLOW_VIEW_SETTINGS                     = 'allow_view_settings';
    const PROJECTS_ALLOW_VIEW_SETTINGS            = self::PROJECTS . '.' . self::ALLOW_VIEW_SETTINGS;


    // User setting.
    const USER_SETTINGS = 'users.settings';
    // User preferences settings.
    const USER_SETTINGS_REVIEW_PREF_SHOW_COMMENTS     = "users.settings.review_preferences.show_comments_in_files";
    const USER_SETTINGS_REVIEW_PREF_SIDE_BY_SIDE      = "users.settings.review_preferences.view_diffs_side_by_side";
    const USER_SETTINGS_REVIEW_PREF_IGNORE_WHITESPACE = "users.settings.review_preferences.ignore_whitespace";
    const USER_SETTINGS_TIME_DISPLAY                  = 'users.settings.time.display';
    // Due to php5.3 not liking constants being put together in a new constant.
    const USER_SETTINGS_REVIEW_PREF_SHOW_SPACE = "users.settings.review_preferences.show_space_and_new_line_characters";

    const SETTINGS = 'settings';

    // Workflow
    // Convenient path prefixes
    const WORKFLOW_ON_SUBMIT_WITH_REVIEW    = 'workflow_rules.on_submit.with_review';
    const WORKFLOW_ON_SUBMIT_WITHOUT_REVIEW = 'workflow_rules.on_submit.without_review';
    const WORKFLOW_END_RULES_UPDATE         = 'workflow_rules.end_rules.update';
    const WORKFLOW_AUTO_APPROVE             = 'workflow_rules.auto_approve';
    // Full workflow paths
    const WORKFLOW_ENABLED                       = 'workflow.enabled';
    const WORKFLOW_ON_SUBMIT_WITH_REVIEW_RULE    = 'workflow_rules.on_submit.with_review.rule';
    const WORKFLOW_GROUP_EXCLUSIONS_RULE         = IWorkflow::WORKFLOW_RULES . '.'
                                                   . IWorkflow::GROUP_EXCLUSIONS . '.'
                                                   . IWorkflow::RULE;
    const WORKFLOW_USER_EXCLUSIONS_RULE          = IWorkflow::WORKFLOW_RULES . '.'
                                                   . IWorkflow::USER_EXCLUSIONS . '.'
                                                   . IWorkflow::RULE;
    const WORKFLOW_GROUP_EXCLUSIONS_MODE         = IWorkflow::WORKFLOW_RULES . '.'
                                                   . IWorkflow::GROUP_EXCLUSIONS . '.'
                                                   . IWorkflow::MODE;
    const WORKFLOW_USER_EXCLUSIONS_MODE          = IWorkflow::WORKFLOW_RULES . '.'
                                                   . IWorkflow::USER_EXCLUSIONS . '.'
                                                   . IWorkflow::MODE;
    const WORKFLOW_ON_SUBMIT_WITHOUT_REVIEW_RULE = 'workflow_rules.on_submit.without_review.rule';
    const WORKFLOW_ON_SUBMIT_WITH_REVIEW_MODE    = 'workflow_rules.on_submit.with_review.mode';
    const WORKFLOW_ON_SUBMIT_WITHOUT_REVIEW_MODE = 'workflow_rules.on_submit.without_review.mode';
    const WORKFLOW_END_RULES_UPDATE_RULE         = 'workflow_rules.end_rules.update.rule';
    const WORKFLOW_END_RULES_UPDATE_MODE         = 'workflow_rules.end_rules.update.mode';
    const WORKFLOW_AUTO_APPROVE_RULE             = 'workflow_rules.auto_approve.rule';
    const WORKFLOW_AUTO_APPROVE_MODE             = 'workflow_rules.auto_approve.mode';
    const WORKFLOW_COUNTED_VOTES_RULE            = 'workflow_rules.counted_votes.rule';
    const WORKFLOW_COUNTED_VOTES_MODE            = 'workflow_rules.counted_votes.mode';

    const SAML_HEADER = 'saml.header';

    const HTTP_STRING             = 'httpString';
    const STRING                  = 'string';
    const INT                     = 'int';
    const INT_OR_HOURS_MINUTES_24 = 'int_or_hours_minutes_24';
    const BOOLEAN                 = 'boolean';
    const ARRAY_OF_STRINGS        = 'arrayOfStrings';
    const SWARM_SETTING           = 'swarmSetting';
    const TYPE                    = 'type';
    const VALID_VALUES            = 'valid_values';
    const ALLOW_NULL              = 'allowNull';
    // For an array of strings we can define metadata that forces string conversion so that an array such
    // as array('1', 2) will be allowed
    const FORCE_STRING = 'force_string';

    // Case sensitivity levels
    const CASE_SENSITIVITY = 'case_sensitivity';
    const CASE_INSENSITIVE = 1;
    const CASE_SENSITIVE   = 0;

    // Some valid value constants
    const VALUE_ANY  = 'any';
    const VALUE_EACH = 'each';

    // Paths listed here are arrays that will have their values replaced rather than merged if specified in the file
    // glob config and also in defaults (glob config defaults override rather than merge)
    const ARRAY_REPLACE_PATHS = [self::REVIEWS_END_STATES];

    // Permitted valid values for the SWARM_SETTING type
    private static $swarmSettings = [
        Setting::FORCED_DISABLED,
        Setting::FORCED_ENABLED,
        Setting::DISABLED,
        Setting::ENABLED
    ];

    // Metadata to define types etc (structure should agree with the configuration array)
    private static $configMetaData = [
        'comments' => [
            'threading' => [
                'max_depth' => [self::TYPE => self::INT]
            ],
            'show_id'                 => [self::TYPE => self::BOOLEAN],
            'notification_delay_time' => [self::TYPE => self::INT]
        ],
        'diffs'   => [
            'max_diffs'             => [self::TYPE => self::INT],
        ],
        'files'   => [
            'max_size'              => [self::TYPE => self::INT],
            'download_timeout'      => [self::TYPE => self::INT]
        ],
        self::REVIEWS => [
            self::SYNC_DESCRIPTIONS => [self::TYPE => self::BOOLEAN],
            self::EXPAND_ALL_FILE_LIMIT => [self::TYPE => self::INT],
            self::FILTERS               => [
                self::RESULT_SORTING  => [self::TYPE => self::BOOLEAN],
                self::DATE_FIELD      =>
                    [self::TYPE => self::STRING, self::VALID_VALUES => ['created', 'updated']]
            ],
            self::EXPAND_GROUP_REVIEWERS          => [self::TYPE => self::BOOLEAN],
            self::DISABLE_TESTS_ON_APPROVE_COMMIT => [self::TYPE => self::BOOLEAN],
            self::DISABLE_APPROVE_WHEN_TASKS_OPEN => [self::TYPE => self::BOOLEAN],
            self::PROCESS_SHELF_DELETE_WHEN       => [
                self::TYPE                 => self::ARRAY_OF_STRINGS,
                self::VALID_VALUES         => [
                    Review::STATE_APPROVED,
                    Review::STATE_APPROVED_COMMIT,
                    Review::STATE_ARCHIVED,
                    Review::STATE_NEEDS_REVIEW,
                    Review::STATE_NEEDS_REVISION,
                    Review::STATE_REJECTED
                ],
                self::CASE_SENSITIVITY => self::CASE_SENSITIVE
            ],
            self::MORE_CONTEXT_LINES       => [self::TYPE => self::INT],
            self::MAX_BOTTOM_CONTEXT_LINES => [self::TYPE => self::INT],
            self::ALLOW_AUTHOR_OBLITERATE  => [self::TYPE => self::BOOLEAN],
            self::DISABLE_SELF_APPROVE     => [self::TYPE => self::BOOLEAN],
            self::DISABLE_COMMIT           => [self::TYPE => self::BOOLEAN],
            self::MODERATOR_APPROVAL       =>
                [self::TYPE  => self::STRING, self::VALID_VALUES => [self::VALUE_ANY, self::VALUE_EACH]],
            self::END_STATES      => [
                self::TYPE                 => self::ARRAY_OF_STRINGS,
                self::VALID_VALUES         => [
                    Review::STATE_APPROVED,
                    Review::STATE_APPROVED_COMMIT,
                    Review::STATE_ARCHIVED,
                    Review::STATE_REJECTED
                ],
                self::CASE_SENSITIVITY => self::CASE_SENSITIVE
            ],
        ],
        self::PROJECTS => [
            self::README_MODE =>  [
                self::TYPE         => self::STRING,
                self::VALID_VALUES => MarkdownSettings::LEGACY_MARKDOWN_OPTIONS
            ],
            self::MAINLINES       => [self::TYPE => self::ARRAY_OF_STRINGS],
            self::MAX_README_SIZE => [self::TYPE => self::INT, self::ALLOW_NULL => true],
            self::FETCH           =>[
                self::MAXIMUM => [self::TYPE => self::INT]
            ],
            self::ADD_ADMIN_ONLY      => [self::TYPE => self::BOOLEAN],
            self::ADD_GROUPS_ONLY     => [self::TYPE => self::BOOLEAN],
            self::ALLOW_VIEW_SETTINGS => [self::TYPE => self::BOOLEAN],
        ],
        'upgrade' => [
            'status_refresh_interval' => [self::TYPE => self::INT],
            'batch_size'              => [self::TYPE => self::INT]
        ],
        'avatars' => [
            'http_url'  => [
                self::TYPE => self::STRING,
                self::CASE_SENSITIVITY => self::CASE_SENSITIVE
            ],
            'https_url' => [
                self::TYPE => self::STRING,
                self::CASE_SENSITIVITY => self::CASE_SENSITIVE
            ]
        ],
        self::SECURITY => [
            'require_login' => [self::TYPE => self::BOOLEAN],
            'prevent_login' => [self::TYPE => self::ARRAY_OF_STRINGS, self::FORCE_STRING => true],
            'https_strict'  => [self::TYPE => self::BOOLEAN],
            self::ADD_PROJECT_ADMIN_ONLY => [self::TYPE => self::BOOLEAN],
            self::ADD_PROJECT_GROUPS     => [self::TYPE => self::BOOLEAN],
            self::EMULATE_IP_PROTECTIONS => [self::TYPE => self::BOOLEAN],
        ],
        self::QUEUE => [
            self::WORKER_CHANGE_SAVE_DELAY => [self::TYPE => self::INT],
            self::PATH                 => [
                self::TYPE => self::STRING,
                self::CASE_SENSITIVITY => self::CASE_SENSITIVE
            ],
            self::WORKERS              => [self::TYPE => self::INT],
            self::WORKER_LIFETIME      => [self::TYPE => self::INT],
            self::WORKER_TASK_TIMEOUT  => [self::TYPE => self::INT],
            self::WORKER_MEMORY_LIMIT  => [
                self::TYPE => self::STRING,
                self::CASE_SENSITIVITY => self::CASE_SENSITIVE
            ],
            self::DISABLE_TRIGGER_DIAGNOSTICS => [self::TYPE => self::BOOLEAN],
        ],
        'translator' => [
            'non_utf8_encodings' => [self::TYPE => self::ARRAY_OF_STRINGS]
        ],
        'users' => [
            'maximum_dashboard_actions' => [self::TYPE => self::INT],
            self::SETTINGS => [
                ReviewPreferences::REVIEW_PREFERENCES => [
                    ReviewPreferences::SHOW_COMMENTS_IN_FILES  => [self::TYPE => self::BOOLEAN],
                    ReviewPreferences::VIEW_DIFFS_SIDE_BY_SIDE => [self::TYPE => self::BOOLEAN],
                    ReviewPreferences::SHOW_SPACE_AND_NEW_LINE => [self::TYPE => self::BOOLEAN],
                    ReviewPreferences::IGNORE_WHITESPACE       => [self::TYPE => self::BOOLEAN]
                ],
                TimePreferences::TIME_PREFERENCES => [
                    TimePreferences::DISPLAY  => [
                        self::TYPE => self::STRING,
                        self::VALID_VALUES => [TimePreferences::TIMEAGO, TimePreferences::TIMESTAMP]
                    ]
                ]
            ],
            'display_fullname' => [self::TYPE => self::BOOLEAN],
        ],
        self::JIRA => [
            self::LINK_TO_JOBS        => [self::TYPE => self::BOOLEAN],
            self::DELAY_JOB_LINKS     => [self::TYPE => self::INT],
            self::MAX_JOB_FIXES       => [self::TYPE => self::INT],
            self::HOST                => [self::TYPE => self::HTTP_STRING],
            self::USER                => [self::TYPE => self::STRING],
            self::PASSWORD            => [self::TYPE => self::STRING, self::CASE_SENSITIVITY => self::CASE_SENSITIVE],
            self::JOB_FIELD           => [self::TYPE => self::STRING, self::CASE_SENSITIVITY => self::CASE_SENSITIVE],
            self::RELATIONSHIP        => [self::TYPE => self::STRING, self::CASE_SENSITIVITY => self::CASE_SENSITIVE],
        ],
        IWorkflow::WORKFLOW => [
            IWorkflow::WORKFLOW_ENABLED => [self::TYPE => self::BOOLEAN]
        ],
        IWorkflow::WORKFLOW_RULES => [
            IWorkflow::ON_SUBMIT => [
                IWorkflow::WITH_REVIEW => [
                    IWorkflow::RULE => [
                        self::TYPE         => self::STRING,
                        self::VALID_VALUES => [IWorkflow::NO_CHECKING, IWorkflow::APPROVED, IWorkflow::STRICT]
                    ],
                    IWorkflow::MODE => [
                        self::TYPE         => self::STRING,
                        self::VALID_VALUES => [IWorkflow::MODE_DEFAULT, IWorkflow::MODE_POLICY]
                    ]
                ],
                IWorkflow::WITHOUT_REVIEW => [
                    IWorkflow::RULE => [
                        self::TYPE         => self::STRING,
                        self::VALID_VALUES => [IWorkflow::NO_CHECKING, IWorkflow::AUTO_CREATE, IWorkflow::REJECT]
                    ],
                    IWorkflow::MODE => [
                        self::TYPE         => self::STRING,
                        self::VALID_VALUES => [IWorkflow::MODE_DEFAULT, IWorkflow::MODE_POLICY]
                    ]
                ]
            ],
            IWorkflow::END_RULES => [
                IWorkflow::UPDATE => [
                    IWorkflow::RULE => [
                        self::TYPE         => self::STRING,
                        self::VALID_VALUES => [IWorkflow::NO_CHECKING, IWorkflow::NO_REVISION]
                    ],
                    IWorkflow::MODE => [
                        self::TYPE         => self::STRING,
                        self::VALID_VALUES => [IWorkflow::MODE_DEFAULT, IWorkflow::MODE_POLICY]
                    ]
                ]
            ],
            IWorkflow::AUTO_APPROVE => [
                IWorkflow::RULE => [
                    self::TYPE         => self::STRING,
                    self::VALID_VALUES => [IWorkflow::NEVER, IWorkflow::VOTES]
                ],
                IWorkflow::MODE => [
                    self::TYPE         => self::STRING,
                    self::VALID_VALUES => [IWorkflow::MODE_DEFAULT, IWorkflow::MODE_POLICY]
                ]
            ],
            IWorkflow::COUNTED_VOTES => [
                IWorkflow::RULE => [
                    self::TYPE         => self::STRING,
                    self::VALID_VALUES => [IWorkflow::ANYONE, IWorkflow::MEMBERS]
                ],
                IWorkflow::MODE => [
                    self::TYPE         => self::STRING,
                    self::VALID_VALUES => [IWorkflow::MODE_DEFAULT, IWorkflow::MODE_POLICY]
                ]
            ],
            IWorkflow::GROUP_EXCLUSIONS => [
                IWorkflow::RULE => [
                    self::TYPE         => self::ARRAY_OF_STRINGS,
                    self::FORCE_STRING => true
                ],
                IWorkflow::MODE => [
                    self::TYPE         => self::STRING,
                    self::VALID_VALUES => [IWorkflow::MODE_DEFAULT, IWorkflow::MODE_POLICY]
                ]
            ],
            IWorkflow::USER_EXCLUSIONS => [
                IWorkflow::RULE => [
                    self::TYPE         => self::ARRAY_OF_STRINGS,
                    self::FORCE_STRING => true
                ],
                IWorkflow::MODE => [
                    self::TYPE         => self::STRING,
                    self::VALID_VALUES => [IWorkflow::MODE_DEFAULT, IWorkflow::MODE_POLICY]
                ]
            ]
        ],
        'saml' => ['header' => [self::TYPE => self::STRING, self::CASE_SENSITIVITY => self::CASE_SENSITIVE]],
        self::MENTIONS => [
            self::USERS_BLACKLIST => [
                self::TYPE => self::ARRAY_OF_STRINGS,
                self::FORCE_STRING => true,
                self::CASE_SENSITIVITY => self::CASE_SENSITIVE
            ],
            self::GROUPS_BLACKLIST => [
                self::TYPE => self::ARRAY_OF_STRINGS,
                self::FORCE_STRING => true,
                self::CASE_SENSITIVITY => self::CASE_SENSITIVE
            ],
            self::MODE => [
                self::TYPE => self::STRING,
                self::VALID_VALUES => ['disabled', 'global', 'projects']
            ]
        ],
        self::P4 => [
            self::SSO_ENABLED => [self::TYPE => self::BOOLEAN],
            P4_SERVER_ID => [self::SSO_ENABLED => [self::TYPE => self::BOOLEAN]]
        ],
        self::ENVIRONMENT => [
            self::LOGOUT_URL => [
                self::TYPE => self::STRING,
                self::CASE_SENSITIVITY => self::CASE_SENSITIVE,
                self::ALLOW_NULL => true
            ],
            self::VENDOR => [
                self::EMOJI_PATH => [
                    self::TYPE => self::STRING,
                    self::CASE_SENSITIVITY => self::CASE_SENSITIVE,
                    self::ALLOW_NULL => true
                ],
            ],
            self::BASE_URL => [
                self::TYPE => self::STRING,
                self::CASE_SENSITIVITY => self::CASE_SENSITIVE,
                self::ALLOW_NULL => true
            ],
            self::ASSET_BASE_PATH => [
                self::TYPE => self::STRING,
                self::CASE_SENSITIVITY => self::CASE_SENSITIVE,
                self::ALLOW_NULL => true
            ]
        ],
        self::LOG => [
            self::REFERENCE_ID => [self::TYPE => self::BOOLEAN],
            self::EVENT_TRACE  => [self::TYPE => self::BOOLEAN],
            self::FILE         => [self::TYPE => self::STRING, self::CASE_SENSITIVITY => self::CASE_SENSITIVE],
            self::PRIORITY     => [self::TYPE => self::INT, self::ALLOW_NULL => true],
        ],
        self::MARKDOWN => [
            self::FILE_EXTENSIONS => [self::TYPE => self::ARRAY_OF_STRINGS, self::FORCE_STRING => true],
            self::MARKDOWN        => [
                self::TYPE => self::STRING, self::VALID_VALUES => MarkdownSettings::MARKDOWN_OPTIONS
            ]
        ],
        self::DEPOT_STORAGE => [
            self::BASE_PATH => [self::TYPE => self::STRING, self::CASE_SENSITIVITY => self::CASE_SENSITIVE]
        ],
        self::REDIS => [
            self::OPTIONS => [
                self::NAMESPACE => [self::TYPE => self::STRING, self::CASE_SENSITIVITY => self::CASE_SENSITIVE]
            ],
            self::POPULATION_LOCK_TIMEOUT => [self::TYPE => self::INT],
            self::ITEMS_BATCH_SIZE        => [self::TYPE => self::INT],
            self::CHECK_INTEGRITY         => [self::TYPE => self::INT_OR_HOURS_MINUTES_24]
        ],
        self::MAIL => [
            self::VALIDATOR => [
                self::OPTIONS => [self::TYPE => self::STRING, self::CASE_SENSITIVITY => self::CASE_SENSITIVE]
            ]
        ]
    ];

    /**
     * Gets a value from the config and checks if it is valid. Simple case at the moment - we may
     * want to enhance to add ranges etc.
     * @param $config array config data
     * @param $path string array path (for nested values use dot notation, for example 'reviews.expand_all_file_limit'.
     * @param null $default default value that will be returned without validation if any problems occur
     * @param null $metaData metadata override
     * @return array|bool|int|mixed|null|string
     * @throws ConfigException
     * @throws \Exception
     */
    public static function getValue($config, $path, $default = null, $metaData = null)
    {
        // This will throw an exception if the value is not set
        $configValue     = null;
        $metaDataElement = null;
        try {
            $configValue = self::getValueFromConfig($config, $path);
            try {
                $metaDataElement = self::getValueFromConfig($metaData ? $metaData : self::$configMetaData, $path);
            } catch (ConfigException $e) {
                // Something is wrong with our metadata, perhaps a path is asked for that we haven't yet catered for.
                // Don't fail because of this.
                return $configValue;
            }
            // This will throw an exception if validation fails
            $configValue = self::validate($configValue, $metaDataElement, $path);
        } catch (ConfigException $ce) {
            if ($default === null || $metaDataElement === null) {
                throw $ce;
            } else {
                $configValue = self::validateDefault($default, $metaDataElement);
            }
        }
        return $configValue;
    }

    /**
     * Make sure that if a default is provided it fits the metadata definition.
     * @param $value
     * @param $metaDataValue
     * @return null
     * @throws ConfigException
     */
    private static function validateDefault($value, $metaDataValue)
    {
        $type    = $metaDataValue[self::TYPE];
        $default = null;
        switch ($type) {
            case self::STRING:
                if (isset($metaDataValue[self::VALID_VALUES]) &&
                    !array_uintersect([$value], $metaDataValue[self::VALID_VALUES], 'strcasecmp')) {
                    throw new ConfigException(
                        "Value '" . $value . "' provided as a default must be one of '"
                        . implode(', ', $metaDataValue[self::VALID_VALUES]) . '\''
                    );
                }
                $default = isset($metaDataValue[self::CASE_SENSITIVITY]) ? $value : strtolower($value);
                break;
            default:
                $default = $value;
                break;
        }
        return $default;
    }

    /**
     * Validates the value.
     * @param mixed     $value          value to validate
     * @param array     $metaDataValue  string meta data
     * @param string    $path           the path
     * @return array|bool|int|null|string
     * @throws ConfigException
     */
    private static function validate($value, array $metaDataValue, $path)
    {
        // Validate all parameters here - currently just type is specified
        $type           = isset($metaDataValue[self::TYPE]) ? $metaDataValue[self::TYPE] : null;
        $convertedValue = null;
        $nullPermitted  =
            $value === null && isset($metaDataValue[self::ALLOW_NULL]) && $metaDataValue[self::ALLOW_NULL];
        $errorMessage   = null;
        switch ($type) {
            case self::INT_OR_HOURS_MINUTES_24:
                if (!is_array($value)) {
                    if (ctype_digit(strval($value))) {
                        $convertedValue = intval($value);
                    } else {
                        // Pick up any negative values before trying as a time
                        if (intval($value) < 0) {
                            $errorMessage = "Value must be >= 0";
                        } else {
                            try {
                                self::validate24HourTime($value);
                                $convertedValue = $value;
                            } catch (InvalidArgumentException $e) {
                                // Leave convertedValue as null
                                $errorMessage = $e->getMessage();
                            }
                        }
                    }
                }
                break;
            case self::INT:
                if (ctype_digit(strval($value))) {
                    $convertedValue = intval($value);
                }
                break;
            case self::BOOLEAN:
                if ($value === true || strcasecmp($value, 'true') == 0) {
                    $convertedValue = true;
                } elseif ($value === false || strcasecmp($value, 'false') == 0) {
                    $convertedValue = false;
                }
                break;
            case self::STRING:
                if (is_string($value)) {
                    if (isset($metaDataValue[self::CASE_SENSITIVITY])
                        && $metaDataValue[self::CASE_SENSITIVITY] === self::CASE_SENSITIVE) {
                        $convertedValue = $value;
                    } else {
                        $convertedValue = strtolower($value);
                    }
                    if (isset($metaDataValue[self::VALID_VALUES]) &&
                        !array_uintersect([$value], $metaDataValue[self::VALID_VALUES], 'strcasecmp')) {
                        // valid values are specified and the value does not match them
                        $convertedValue = null;
                    }
                }
                break;
            case self::ARRAY_OF_STRINGS:
                // We support setting a single string value that we will convert to an array
                $value = is_string($value) ? (array) $value : $value;
                if (is_array($value)) {
                    $convertedValue = [];
                    foreach ($value as $arrayValue) {
                        if (isset($metaDataValue[self::FORCE_STRING])) {
                            $arrayValue = (string)$arrayValue;
                        }
                        if (is_string($arrayValue)) {
                            // Check if the case sensitivity enabled.
                            if (isset($metaDataValue[self::CASE_SENSITIVITY])
                                && $metaDataValue[self::CASE_SENSITIVITY] === self::CASE_SENSITIVE) {
                                $convertedValue[] = $arrayValue;
                                // Now also compare the array in case sensitive mode.
                                if (isset($metaDataValue[self::VALID_VALUES]) &&
                                    !array_intersect([$arrayValue], $metaDataValue[self::VALID_VALUES])) {
                                    // valid values are specified and the value does not match them
                                    $convertedValue = null;
                                    break;
                                }
                            } else {
                                $convertedValue[] = strtolower($arrayValue);
                                if (isset($metaDataValue[self::VALID_VALUES]) &&
                                    !array_uintersect(
                                        [$arrayValue],
                                        $metaDataValue[self::VALID_VALUES],
                                        'strcasecmp'
                                    )
                                ) {
                                    // valid values are specified and the value does not match them
                                    $convertedValue = null;
                                    break;
                                }
                            }
                        } else {
                            $convertedValue = null;
                            break;
                        }
                    }
                }
                break;
            case self::HTTP_STRING:
                if (is_string($value)) {
                    $convertedValue = rtrim($value, '/');
                    if ($convertedValue && strpos(strtolower($convertedValue), 'http') !== 0) {
                        $convertedValue = 'http://' . $convertedValue;
                    }
                }
                break;
            case self::SWARM_SETTING:
                if (array_uintersect([$value], self::$swarmSettings, 'strcasecmp')) {
                    // SWARM_SETTING has its own set of permitted values in configuration
                    // that are case sensitive
                    $convertedValue = $value;
                } else {
                    $convertedValue = null;
                }
                break;

            case null:
                $convertedValue = $value;
                break;
        }
        if ($convertedValue === null && !$nullPermitted) {
            throw new ConfigException(
                "Value '" . (is_array($value) ? var_export($value, true) : $value) .
                "' at path '" . $path . "' is invalid" . ($errorMessage ? ". $errorMessage" : '')
            );
        }
        return $convertedValue;
    }

    /**
     * Iterates through the configuration to find a value.
     * @param $config array config
     * @param $path string array path (for nested values use dot notation, for example 'reviews.expand_all_file_limit'.
     * @return mixed
     * @throws ConfigException if the path does not exist or the configuration being searched is not an array.
     */
    private static function getValueFromConfig($config, $path)
    {
        $ref  = &$config;
        $keys = explode('.', $path);
        foreach ($keys as $idx => $key) {
            if (!is_array($ref)) {
                throw new ConfigException('Configuration is not an array');
            }
            if (!array_key_exists($key, $ref)) {
                throw new ConfigException("Path '" . $path . "' does not exist", ConfigException::PATH_DOES_NOT_EXIST);
            }
            $ref = &$ref[$key];
        }
        return $ref;
    }

    /**
     * Updates the merged config by overriding it with a value from the config.php file if it was set.
     * This gives the ability to replace values in merged config rather than merging them in. For example
     * Zend merges arrays so default settings of [1, 2, 3] in a module.config.php and [4] in the glob file
     * result in a merged value of [1, 2, 3, 4] when in some cases we just want [4]. Any paths that are present
     * in self::ARRAY_REPLACE_PATHS will be treated as replacements rather than merges.
     * @param ConfigMergerInterface $configListener the config listener
     * @param string                $globConfigPath config glob path
     * @param array|null            $paths          optional paths to use in place of self::ARRAY_REPLACE_PATHS
     * @throws ReflectionException
     * @throws Exception if the path does not exist or the configuration being searched is not an array.
     */
    public static function updateMerged(ConfigMergerInterface $configListener, $globConfigPath, $paths = null)
    {
        // Using reflection is not ideal but overriding the ConfigListener to access values
        // has not proved possible with the current version of Zend
        $reflection = new ReflectionClass('\Zend\ModuleManager\Listener\ConfigListener');
        $var        = $reflection->getProperty('configs');
        $var->setAccessible(true);
        $configs = $var->getValue($configListener);
        $paths   = $paths ? $paths : self::ARRAY_REPLACE_PATHS;
        if (isset($configs[$globConfigPath])) {
            $mergedConfig = $configListener->getMergedConfig(false);
            // If path value is set in the config.php file then use that rather than the value
            // created from merging Module/module.config.php with the glob file
            $globConfig = $configs[$globConfigPath];
            foreach ($paths as $path) {
                try {
                    $value = self::getValueFromConfig($globConfig, $path);
                    if ($value && !empty($value)) {
                        $keys = explode('.', $path);
                        $temp = &$mergedConfig;
                        foreach ($keys as $key) {
                            $temp = &$temp[$key];
                        }
                        $temp = $value;
                    }
                } catch (ConfigException $e) {
                    // Path was not set in the glob file, will fall back to defaults
                }
            }
            $configListener->setMergedConfig($mergedConfig);
        }
    }

    /**
     * Validates that value is in 24 hour time format
     * @param string|null $value the value
     * @return string the value
     * @throws InvalidArgumentException if the value is not in the correct format
     */
    public static function validate24HourTime($value)
    {
        $matches = [];
        preg_match(self::HOURS_MINUTES_24_PATTERN, $value, $matches);
        if ($matches) {
            return $value;
        }
        throw new InvalidArgumentException("Value does not match 24 hour pattern HH:ii");
    }

    /**
     * Here we want to change the default behaviour whereby Zend merges arrays together rather than replaces when
     * values are provided in defaults. This function is attached as a listener in the Application Module to
     * be triggered on the ModuleEvent::EVENT_MERGE_CONFIG event
     * @param ModuleEvent $e
     * @throws \ReflectionException
     */
    public function mergeConfig(ModuleEvent $e)
    {
        if (defined('BASE_DATA_PATH')) {
            self::updateMerged(
                $e->getConfigListener(),
                realpath(BASE_DATA_PATH . '/config.php')
            );
        }
    }
}
